//===================================CISCKER====================================
//=============================Instruction Decoder==============================
//------------------------------------------------------------------------------
//File Name:	instruction_decoder.v
//Purpose:		Top decoding module, instantiates ALU operation, branch, u-Instruction and main decoder
//Parameters:	None
//------------------------------------------------------------------------------
//=====================Alfredo Gualdron - Jose P. Pinilla=======================
//=================================Semillero ADT================================
//================================UPB Bucaramanga===============================
//Website:	http://semilleroadt.upbbga.edu.co/xiscker
//E-mail:	semilleroadt@gmail.com
//------------------------------------------------------------------------------
module instruction_decoder(
	//inputs
	input [7:0]opcode,
	input load,
	input clk_0,
	input clk_90,
	input [5:0]CCR,
	//outputs
	output [5:0] alu_op,
	output [4:0] alu_operands,
	output reg[3:0] adder_operands,
	output reg[7:0]control=0,
	output wd_src,
	output branch,
	output adder_alu,
	output acca_src,
	//data destination
	output ccr_en,
	output sp_en,
	output ix_en,
	output acca_en,
	output accb_en,
	output hidden_en,
	output reg final=0
	);

///////////////////////////////////////////////////////////////////////////////
//u-Instructions
///////////////////////////////////////////////////////////////////////////////
parameter p	 =	2'd0;	 //Program fetch 
parameter r	 =	2'd1;	 //Read byte 
parameter d	 =	2'd2;	 //Dummy 
parameter w	 =	2'd3;	 //Write

///////////////////////////////////////////////////////////////////////////////
//MAIN DECODER
///////////////////////////////////////////////////////////////////////////////
wire [35:0] next_queue;
wire [8:0] operands;
wire [4:0] destination;


main_decoder decoder(
			//inputs
			.clk_0(clk_0),
			.final_90(final),
			.opcode(opcode),
			//outputs
			.next_queue(next_queue),
			.adder_alu_operands(operands), 	//{adder_A(2),adder_B(2),ALU_A(3),ALU_B(2)}
			.result_destination(destination)	//{SP,IX,ACCA,ACCB,hidden}
			);

///////////////////////////////////////////////////////////////////////////////
//EXECUTION QUEUE
///////////////////////////////////////////////////////////////////////////////
parameter addressing_state = 36'b110100;
reg [35:0]queue = addressing_state;
/////////////////

wire lookahead;
wire operation;
wire comp; //Complement
wire [1:0]u_inst;

always@(posedge clk_0)
	if (load)
		queue <= addressing_state;
	else if(final)
		queue <= next_queue;
	else
		queue <= queue >> 6;

assign {lookahead,operation,comp,u_inst} = queue[4:0];

always@(posedge clk_90) //atrazo de la carga de la siguiente fila
	final <= queue[5];
	
///////////////////////////////////////////////////////////////////////////////
//ALU OPERATION DECODER
///////////////////////////////////////////////////////////////////////////////

alu_operation_decoder alu_op_decoder(
			//in
			.clk_0(clk_0),
			.final(final),
			.opcode(opcode),
			.operation(operation),
			.carry(CCR[0]),
			//out
			.alu_op(alu_op)
			);

//////////////////////////////////////////////////////////////////////////////////
//OPERATION CONTROL SIGNALS
//////////////////////////////////////////////////////////////////////////////////

assign adder_alu = operation & (sp_en | ix_en);				//ALU operation to be registered in SP or IX
		
assign wd_src = (u_inst==w) & comp;							//MS or LS data to be written
		
assign acca_src = acca_en & ~accb_en;						//Destination is ACCA
		
wire ix_plus = (opcode[7:4]==4'h7);							//Set for IX+ adressing instructions
		
wire stack =	(opcode==8'h26)|(opcode==8'h27)|			//Set for Push and Pull Instructions
				(opcode==8'h28)|(opcode==8'h29)|			//PSHA,PSHB,PULA,PULB,PSHX,PULX
				(opcode==8'h2A)|(opcode==8'h2B)|
				(opcode==8'h2F)|(opcode==8'h2D); 			//SWI-RTI

wire rti_instr = (opcode==8'h2D);

wire MOV = (opcode==8'h5E)|(opcode==8'h6E)&(operation); 	//MOV needs ALU operands to change from MM/hidden to 00/hidden

//////////////////////////////////////////////////////////////////////////////
//BRANCH DECODER
///////////////////////////////////////////////////////////////////////////////

branch_decoder branch_inst(
			//inputs
			.clk_0(clk_0),
			.final(final),
			.opcode(opcode),
			.CCR(CCR),
			//output
			.branch(branch)
			);	

wire pa_d = branch | comp;		//Branch condition for u-Instruction d "pass_adder"
wire pc_d = branch & (~comp);	//Branch condition for u-Instruction d "pc_en"
			
//////////////////////////////////////////////////////////////////////////////////////////////////
//u-Instruction DECODER
//////////////////////////////////////////////////////////////////////////////////////////////////

always @ (*)
	case (u_inst)
//										opcode_en:																
//										|	pass_adder:
//										|	|	pc_en:												
//										|	|	|	  adder_cin:
//										|	|	|	  |		h_reset:
//										|	|	|	  |		|	mah_en:
//										|	|	|	  |		|	|	we:
//										|	|	|	  |		|	|	|	mdr_en:
		p:			control =	{lookahead,1'b1,1'b1,1'b1,comp,1'b1,1'b0,1'b1};
		r:			control =	{lookahead,1'b1,comp,1'b1,1'b0,1'b0,1'b0,1'b1};
		d:			control =	{lookahead,pa_d,pc_d,1'b0,1'b1,1'b0,1'b0,1'b1};
		w:			control =	{lookahead,1'b1,comp,1'b1,1'b0,1'b0,1'b1,1'b0};
	endcase

//////////////////////////////////////////////////////////////////////////////////////////////////
//Data destination
//////////////////////////////////////////////////////////////////////////////////////////////////

//PSH or PUL instructions don't update CCR
assign ccr_en 		=	stack ? (opcode==8'h2D)&acca_en : operation;
//Destination decoder
assign sp_en 		=	stack ?   ~operation : (destination[4] & operation);
assign ix_en 		=	ix_plus ? (destination[3] & (u_inst==d)) : (destination[3] & operation);
assign acca_en 		=	destination[2] & operation;
assign accb_en 		=	destination[1] & operation;
assign hidden_en 	=	destination[0] & ~operation;

//////////////////////////////////////////////////////////////////////////////////////////////////
//OPERANDS
//////////////////////////////////////////////////////////////////////////////////////////////////

//ALU Operands
assign alu_operands = MOV ? 5'b111_10 : operands[4:0]; //MOV needs ALU operands to change from MM/hidden to 00/hidden

//Adder Operands Decoder
always @ (*)
case (u_inst)
	p:	adder_operands = 4'h0;
	
	r,d:case (opcode)
			8'h2C:	if (operation)		//RTS
						adder_operands = operands[8:5];	//SP+00
					else
						adder_operands = 4'b1100;	//MM+00
			
			8'h5A,8'h5B,8'h6A, 8'h6B:	//BRCLR,BRSET IX&EXT
					if (comp)
						adder_operands = operands[8:5];
					else
						adder_operands = 4'b0011;	//PC+MM
			
			8'h5E, 8'h6E: 				//MOV IX, MOV DIR
					if(comp)
						adder_operands = operands[8:5];
					else
						adder_operands = 4'b0000;	//PC+00
			
			default:	adder_operands = operands[8:5];
		endcase
				
	w:	begin		
		if (comp)
			if (opcode[3:0]==4'h9)
				case (opcode[7:4])		//BSR, JSR DIR, JSR IX, JSR EXT
					4'hC:	adder_operands = 4'b0011;	//PC+MM
					4'hD:	adder_operands = 4'b1100;	//MM+00
					4'hE:	adder_operands = 4'b1011;	//IX+MM
					4'hF:	adder_operands = 4'b1100;	//MM+00
					default:adder_operands = 4'b0000;	//PC+00
				endcase
			else
				adder_operands = 4'h0;
		else
			adder_operands = operands[8:5];
		end
	
endcase

endmodule
